/* OS- and machine-dependent stuff for IBM-PC running MS-DOS */
#ifdef __TURBOC__
#pragma inline
#endif

#include <stdio.h>

#ifdef MSC
#include <process.h>
#include <fcntl.h>
#include <io.h>
#endif

#if defined(__TURBOC__)
#include <dos.h>
#include <dir.h>
#include <sys/stat.h>
#include <string.h>
#include <process.h>
#include <fcntl.h>
#endif

#include "global.h"
#include "mbuf.h"
#include "internet.h"
#include "iface.h"
#include "pc.h"
#include "cmdparse.h"

#if defined(__TURBOC__)
extern unsigned char model; /* set in maintc.c 0 = P+; 1 = P; 2 = PC */
#endif

/* This flag is set by setirq() if IRQ 8-15 is used, indicating
 * that the machine is a PC/AT with a second 8259 interrupt controller.
 * If this flag is set, the interrupt return code in pcgen.asm will
 * send an End of Interrupt command to the second 8259 as well as the
 * first.
 */
char isat;	

/* Interface list header */
struct interface *ifaces;

/* Aztec memory allocation control */
int _STKLOW = 0;	/* Stack above heap */
int _STKSIZ = 16384/16;	/* 16K stack -- overridden in grabcore */
int _HEAPSIZ = 4096/16;	/* Isn't really used */
int _STKRED = 4096;	/* Stack red zone in bytes -- this really matters */

char ttbuf[BUFSIZ];

/* Stolen from Nansi.sys' SETRAW.MSC program which was stolen in turn from
 * the Hack sources.  Also, second hand thanks to Mark Zbikowski
 * (markz@microsoft.UUCP)
 */

#define IOCTL           0x44		/* Dos Fn for I/O Control of Devices */
#define GETBITS         0		/*    Get Device Information         */
#define SETBITS         1		/*    Set Device Information	     */

#define DEVICE          0x80		/* 0=attached to file, 1=to device   */
#define RAW             0x20		/* Binary bit for Devices	     */

static unsigned old_stdin, old_stdout, ioctl();

set_raw()
{
        old_stdin = ioctl(fileno(stdin), GETBITS, 0);
        old_stdout = ioctl(fileno(stdout), GETBITS, 0);
        if (old_stdin & DEVICE)
                ioctl(fileno(stdin), SETBITS, old_stdin | RAW);
        if (old_stdout & DEVICE) {
                ioctl(fileno(stdout), SETBITS, old_stdout | RAW);
#ifdef MSC
		/* Disable CRLF translations */
		(void)setmode(fileno(stdout), O_BINARY) ;
#endif
	}
}
restore_raw()
{
        if (old_stdin)
                (void) ioctl(fileno(stdin), SETBITS, old_stdin);
        if (old_stdout) {
                (void) ioctl(fileno(stdout), SETBITS, old_stdout);
#ifdef MSC
		/* Re-enable CRLF translations */
		(void)setmode(fileno(stdout), O_TEXT) ;
#endif
	}
}
static unsigned
ioctl(handle, mode, setvalue)
unsigned setvalue;
{
	unsigned short ax,bx,cx,dx;
 
        ax = (IOCTL<<8) + mode;
        bx = handle;
	cx = 0;			/* Not used */
        dx = setvalue;		/* Zero out dh */
        int21h(&ax,&bx,&cx,&dx);
        return dx;
}

/* Called at startup time to set up console I/O, memory heap */
ioinit()
{
	unsigned grabcore();
 	/* Save these two file table entries for something more useful */
 	fclose(stdaux);
#if defined(__TURBOC__)
 	fclose(stdprn);
#else
 	fclose(stdprt);
#endif

	/* Interrupts use a special stack deep in data space.
	 * Calls to sbrk() (invoked by malloc when it needs more memory
	 * from the system) at interrupt time will fail because sbrk()
	 * will think that the stack has overwritten the heap. So
	 * grab all the memory we can now for the heap so that malloc
	 * won't have to call sbrk and alloc_mbuf() won't fail unnecessarily
	 * at interrupt time.
	 */
	grabcore();

	setbuf(stdout, ttbuf);
#ifdef billdo_not
 	/* Put display in raw mode. Note that this breaks tab expansion,
 	 * so you need to run NANSI.SYS or equivalent.
 	 */
	set_raw();
#endif
}
/* Called just before exiting to restore console state */
iostop()
{
	/* put stdout in cooked mode */
	setbuf(stdout,NULLCHAR);
	restore_raw();
	cooked();

	while(ifaces != NULLIF){
		if(ifaces->stop != NULLFP)
			(*ifaces->stop)(ifaces);
		ifaces = ifaces->next;
	}
}
/* Spawn subshell */
doshell(argc,argv)
{
	char *command,*getenv();
	int ret;

	/* put stdout in cooked mode */
	setbuf(stdout,NULLCHAR);
	restore_raw();
	cooked();

	if((command = getenv("COMSPEC")) == NULLCHAR)
		command = "/COMMAND.COM";
#if defined(__TURBOC__) || defined(MSC)
	ret = spawnl(P_WAIT,command,command,NULLCHAR);
#if defined(PLUS)
/*
 * set: cursor to block, attributes off, keyboard to ALT mode,
 *      transmit functions off, use HP fonts
 */
	printf(/* "\033*dK"	   cursor to block		*/
	          "\033&d@"	/* display attributes off  	*/
	       /* "\033&k1\\"	   KBD to ALT mode		*/
	       /* "\033&s1A"	   transmit functions on	*/
	          "\033[11m");	/* use ALT fonts		*/
#endif
#else
  	ret = fexecl(command,command,NULLCHAR);
  	if(ret != -1)
  		ret = wait();
#endif

	/* put stdout in raw mode */
	setbuf(stdout, ttbuf);
	set_raw();

	return ret;
}

/* checks the time then ticks and updates ISS */
static short clockstart =0;
static unsigned clkval = 0;
/* checks the time then ticks and updates ISS */
void
check_time()
{
	int32 iss();
	short ntime;

#if !defined(NOMAD)   /* not compiling for a HP-110 */
	if(!clockstart){
		/* Executed only once */

#if defined(PLUS)
		clkval = peek(0,0x472);		/* compiling for a HP-110+  */
#endif
#if !defined(PLUS) && defined(__TURBOC__)
		clkval = peek(0,0x46c);
#endif
#if !defined(PLUS) && !defined(__TURBOC__)
		clkval = peekw(0x6c,0x40);
#endif

		clockstart = 1;
		return;
	}
	/* Read the low order word of the BIOS tick counter directly. The
	 * INT 1Ah call isn't used because it stupidly clears the
	 * "midnight passed" flag and we'd have to update the date ourselves.
	 * (See Norton, p222).
	 * 
 	 * The PC's time-of-day handling is a real crock of shit. Why not a
	 * nice simple long binary count from a fixed UTC epoch, as in UNIX??
	 */

#if defined(PLUS)
	ntime = peek(0,0x472);
#endif
#if !defined(PLUS) && defined(__TURBOC__)
	ntime = peek(0,0x46c);
#endif
#if !defined(PLUS) && !defined(__TURBOC__)
	ntime = peekw(0x6c,0x40);
#endif

	while(ntime != clkval){	/* Handle possibility of several missed ticks */
		clkval++;
#else /* compiling for a HP-110 */
	if(clkval != clksec()) {
		clkval = clksec();
#endif /* NOMAD */
		icmpclk();	/* Call this one before tick */
		tick();
		(void)iss();
	}
}

/* Read characters from the keyboard, translating them to "real" ASCII
 * If none are ready, return the -1 from kbraw()
 */
int
kbread()
{
	int kbraw(),c;

	if((c = kbraw()) == 0){
		/* Lead-in to a special char */
		c = kbread();
		switch(c){
		case 3:		/* NULL (bizzare!) */
			c = 0;
			break;
#if defined(PLUS)
		case 66:	/* F- 8 too */
#endif
		case 68:	/* F-10 key (used as command-mode escape) */
			c = -2;
			break;
		case 83:	/* DEL key */
			c = 0x7f;
			break;
		default:	/* Dunno what it is */
			c = -1;
		}
	}
/*
 * allow either HP or ALT mode f8 to be command-mode escape for Portables
 */
#if defined(PLUS)
	if(c==27 && kbhit()) {
		if((c = kbraw()) == 119) return -2;
		return -1;
	}
#endif
	return c;
}
/* Reset the CPU, reboot DOS */
sysreset()
{
	int16 regs[8];	/* Not really needed */
#if defined(__TURBOC__)
	asm db 9ah,0ffh,0ffh,0,0;
#else
	farcall(0x0,0xffff,regs,regs);
#endif
}


#if !defined(__TURBOC__)
#define	CTLZ	26
/* Special version of aputc() (used by putchar and printf) that filters
 * out nasty characters that screw up the DDOS and ANSI terminal drivers
 */
aputc(c,file)
char c;
FILE *file;
{
 	/* Nulls get displayed as spaces by ansi.sys (wrong)
 	 * ^Z's seem to hang the DoubleDos and DesqView screen drivers
 	 */
 	if((c == '\0' || c == CTLZ) && file == stdout)
 		return c;
 	/* Do end-of-line translations */
 	if(c == '\n')
 		putc('\r',file);
 	return putc(c,file);
}
#endif /* !defined(__TURBOC__) */

/* Install hardware interrupt handler.
 * Takes IRQ numbers from 0-7 (0-15 on AT) and maps to actual 8086/286 vectors
 * Note that bus line IRQ2 maps to IRQ9 on the AT
 */
setirq(irq,handler)
unsigned irq;
int (*handler)();
{
	/* Set interrupt vector */
	if(irq < 8){
		setvect(8+irq,handler);
	} else if(irq < 16){
		isat = 1;
		setvect(0x70 + irq - 8,handler);
	} else {
		return -1;
	}
	return 0;
}

/* Return pointer to hardware interrupt handler.
 * Takes IRQ numbers from 0-7 (0-15 on AT) and maps to actual 8086/286 vectors
 */
int (*getirq(irq))()
{
#if defined(__TURBOC__)
void	interrupt 	(* _Cdecl getvect(int interruptno)) ();
#else
	int (*getvect())();
#endif

	/* Set interrupt vector */
	if(irq < 8){
		return getvect(8+irq);
	} else if(irq < 16){
		return getvect(0x70 + irq - 8);
	} else {
		return NULLFP;
	}
}

/* Disable hardware interrupt */
maskoff(irq)
unsigned irq;
{
	if(irq < 8){
		setbit(0x21,(char)(1<<irq));
	} else if(irq < 16){
		irq -= 8;
		setbit(0xa1,(char)(1<<irq));
	} else {
		return -1;
	}
	return 0;
}

/* Enable hardware interrupt */
maskon(irq)
unsigned irq;
{
	if(irq < 8){
		clrbit(0x21,(char)(1<<irq));
	} else if(irq < 16){
		irq -= 8;
		clrbit(0xa1,(char)(1<<irq));
	} else {
		return -1;
	}
	return 0;
}

#if defined(__TURBOC__)
/*
 * for interrupt stack
 */
unsigned char intstk[1520];
unsigned int sssave, spsave;
static int break_state;
char ttbuf[BUFSIZ];

doret()
{
	asm cmp isat,1;
	asm mov al,0bh;  /* read in service register from */
	asm out 0a0h,al; /* secondary 8259                */
	asm nop;	 /* settling delay                */
	asm nop;
	asm nop;
	asm in al,0a0h;	 /* get it                        */
	asm or al,al;	 /* Any bits set?                 */
	asm jz notat;    /* Nope, not a secondary intr    */
	asm mov al,20h;  /* Get EOI instruction           */
	asm out 0a0h,al; /* Secondary 8259 (PC/AT only    */
notat:;
	asm mov al,20h;
	asm out 20h,al;
}

#endif

#if defined(__TURBOC__)
/*
 * restore - restore interrupt state: 0 = off, nonzero = on
 */
void restore(istate)
char istate;
{
	asm test byte ptr istate,0ffh
	asm jz ioff
	asm sti
ioff:
} /* restore */
#endif /* defined(__TURBOC__) */

#if defined(__TURBOC__)
/*
 * Halt until an interrupt occurs, then return
 */
void eihalt()
{
	asm sti
	asm hlt
} /* eihalt */
#endif /* defined(__TURBOC__) */

#if defined(__TURBOC__)
/*
 * get raw keyboard state
 */
int kbraw()
{
	asm mov ah,06h	/* Direct Console I/O  */
	asm mov dl,0ffh	/* Read from keyboard  */
	asm int 021h	/* Call DOS  */
	asm jz nochar 	/* zero flag set -> no character ready  */
	return(_AL); 	/* valid char is 0-255  */
nochar:
	return(-1);
}
#endif /* defined(__TURBOC__) */

#if defined(__TURBOC__)
/*
 * calculate free stuff for dirutil
 */
void int21h(a,b,c,d)
unsigned *a, *b, *c, *d;
{
	unsigned i;

	_AX= (*a);
	_DX= (*d);	/* default drive == 0 */
	__int__(0x21);
	i=_BX;
	*a=_AX;
	*b=i;
	*c=_CX;
	*d=_DX;
} /* isfree */
#endif /* defined(__TURBOC__) */

#if defined(__TURBOC__)
/*
 * Internet checksum subroutine
 * Compute 1's-complement sum of data buffer
 */
unsigned short lcsum(buf, cnt)
unsigned short *buf, cnt;
{
	asm 	push	si	
#if defined(__COMPACT__) || defined(__LARGE__) || defined(__HUGE__)
	asm 	push 	ds
	asm	lds	si, buf	/* ds:si = buf */
#else
	asm	mov	si, buf
#endif
	asm	mov	cx,cnt		/* cx = cnt		*/
	asm	clc			/* clear carry		*/
	asm	cld			/* autoincrement si	*/
	asm	mov	dx,0		/* clear accumulated sum*/
lloop:	asm	lodsw			/* do { ax = *si++	*/
	asm	adc	dx,ax		/* dx += ax + cy	*/
	asm	loop	lloop		/* } while(--cx != 0);	*/
	asm	adc	dx,0		/* get last carries	*/
	asm	adc	dx,0
	asm	mov	ax,dx		/* result into ax	*/
	asm	xchg	al,ah	/* byte swap result (8088 is little-endian) */
#if defined(__COMPACT__) || defined(__LARGE__) || defined(__HUGE__)
	asm 	pop 	ds
#endif
	asm 	pop	si	
} /* lcsum */
#endif /* defined(__TURBOC__) */

#if defined(__TURBOC__) && !defined(PLUS)
/*
 * Relinquish processor so other task can run
 * multitasker types
 */
unsigned char mtasker;
#define NONE		0
#define DOUBLEDOS 	1
#define DESQVIEW	2
void giveup()
{
	if(mtasker==DOUBLEDOS) {
		_AL=2;
		_AH=0xee;
		__int__(0x21);
	}
	if(mtasker==DESQVIEW) {
		_AX=0x1000;
		__int__(0x15);
	}
} /* giveup */
#endif /* defined(__TURBOC__) && !defined(PLUS) */

#if defined(__TURBOC__) && !defined(PLUS)
/*
 * check for a multitasker running
 */
void chktasker()
{
/*
 * test doubledos
 */
	asm 	mov 	ah, 0e4h
	asm 	int 	21h
	asm 	cmp 	al,1
	asm 	jz	isdos
	asm 	cmp 	al,2
	asm 	jnz 	test_desq
isdos:  mtasker=DOUBLEDOS;
	return;
/*
 * test for desqview
 */
 test_desq:
	asm	mov	ax, 2b01h
	asm	mov	cx, 4445h
	asm	mov	dx, 5351h
	asm	int	21h
	asm	cmp	al, 0ffh
	asm	jnz	isdesq
	return;
isdesq: mtasker=DESQVIEW;
} /* chktasker */
#endif /* defined(__TURBOC__) && !defined(PLUS) */

#if defined(__TURBOC__) && defined(NOMAD)
/*
 * returns the second value of the clock
 */
int clksec()
{
	_AH=0x2c;
	__int__(0x21);
	_AL=_DH;
	_AH=0;
}
#endif /* defined(__TURBOC__) && defined(NOMAD) */
